/*************************************************************************
 This class enables C++ code to catch Win32 structured exceptions. Its 
 static member function handle() can be used to specify whether or not 
 structured exceptions should be handled on a per-thread basis. Where 
 they are handled a Win32Exception is thrown. The class includes a 
 toString() function which details the exception. Several other member 
 functions exist to determine the exception code, the exception type, 
 the exception address, the name of the module containing the exception 
 address and the offset from the start of that module to the exception
 address.
*************************************************************************/
#include <windows.h>
#include "Win32Exception.h"

/*************************************************************************
 For each thread we save the address of the previous SE handler 
 translator. We make no attempt to maintain a chain of previous handlers.
 We assume that we want to translate structured exceptions to C++ 
 exceptions or we want the default behavior.
*************************************************************************/
int Win32Exception::tlsIndex = -1;

/*************************************************************************
 Old habits die hard deaths
*************************************************************************/
char Win32Exception::fmt[] = "%08.08X";

/*************************************************************************
 The constructor saves the exception code and the address where the
 structured exception was encountered.
*************************************************************************/
Win32Exception::Win32Exception(DWORD dwCode, LPEXCEPTION_POINTERS pexp)
{
 dwExcCode = dwCode; 
 dwAddress = reinterpret_cast<DWORD>(pexp->ExceptionRecord->ExceptionAddress);
}

/*************************************************************************
 The constructor saves the exception code and the address where the
 structured exception was encountered.
*************************************************************************/
Win32Exception::Win32Exception(DWORD dwCode, DWORD dwLocation)
{
 dwExcCode = dwCode; 
 dwAddress = dwLocation;
}

/*************************************************************************
 The destructor doesn't do anything
*************************************************************************/
Win32Exception::~Win32Exception(){}

/*************************************************************************
 This is the exception handler translator. We create an instance of the
 Win32Exception class and then throw it.
*************************************************************************/
void Win32Exception::sehTranslator(UINT u, LPEXCEPTION_POINTERS pexp)
{
 Win32Exception exc(u, pexp);
 throw exc;
}

/*************************************************************************
 The function enables the translation of structured exceptions to C++
 exceptions. 
*************************************************************************/
bool Win32Exception::enable(int nIndex)
{
 bool                    ok;
 _se_translator_function oldTrans;

 if ( tlsIndex == -1 )
  tlsIndex = nIndex;

 if ( tlsIndex == -1 )
  ok = false;

 else
 {
  oldTrans = _set_se_translator(&Win32Exception::sehTranslator);
  TlsSetValue(tlsIndex, oldTrans);

  ok = true;
 }

 return ok;   
}

/*************************************************************************
 The function enables the translation of structured exceptions to C++
 exceptions. 
*************************************************************************/
bool Win32Exception::enable(void)
{
 if ( tlsIndex == -1 )
  tlsIndex = TlsAlloc();

 return Win32Exception::enable(tlsIndex); 
}

/*************************************************************************
 The function disables the translation of structured exceptions to C++
 exceptions. 
*************************************************************************/
void Win32Exception::disable(void)
{
 _se_translator_function oldTrans;

 if ( tlsIndex != -1 )
 {
  oldTrans = (_se_translator_function) TlsGetValue(tlsIndex);
  _set_se_translator(oldTrans);

  TlsSetValue(tlsIndex, 0);
 }

 return;   
}

/*************************************************************************
 The function returns the address where the exception occurred.
*************************************************************************/
DWORD Win32Exception::getCode() const
{
 return dwExcCode;
}

/*************************************************************************
 The function returns the address where the exception occurred.
*************************************************************************/
DWORD Win32Exception::getAddress() const
{
 return dwAddress;
}

/*************************************************************************
 The function returns the type of exception that occurred.
*************************************************************************/
std::string Win32Exception::getType() const
{
 std::string str;

 switch ( dwExcCode )
 {
  case EXCEPTION_ACCESS_VIOLATION:

       str = "Access Violation";
       break;

  case EXCEPTION_DATATYPE_MISALIGNMENT:

       str = "Datatype Misalignment";
       break;

  case EXCEPTION_BREAKPOINT:

       str = "Breakpoint";
       break;

  case EXCEPTION_SINGLE_STEP:

       str = "Single Step";
       break;

  case EXCEPTION_ARRAY_BOUNDS_EXCEEDED:

       str = "Array Bounds Exceeded"; 
       break;

  case EXCEPTION_FLT_DENORMAL_OPERAND:

       str = "Denormal Floating Point Operand";
       break;

  case EXCEPTION_FLT_DIVIDE_BY_ZERO:

       str = "Floating Point Divide by Zero";
       break;

  case EXCEPTION_FLT_INEXACT_RESULT:

       str = "Floating Point Inexact Result";
       break;

  case EXCEPTION_FLT_INVALID_OPERATION:

       str = "Floating Point Invalid Operation";
       break; 

  case EXCEPTION_FLT_OVERFLOW:

       str = "Floating Point Overflow";
       break;

  case EXCEPTION_FLT_STACK_CHECK:

       str = "Floating Point Stack Check";
       break;

  case EXCEPTION_FLT_UNDERFLOW:

       str = "Floating Point Underflow";
       break;

  case EXCEPTION_INT_DIVIDE_BY_ZERO:

       str = "Integer Division by Zero";
       break;

  case EXCEPTION_INT_OVERFLOW:

       str = "Integer Overflow";
       break;

  case EXCEPTION_PRIV_INSTRUCTION:

       str = "Privileged Instruction";
       break;

  case EXCEPTION_IN_PAGE_ERROR:

       str = "In Page Error";
       break;

  case EXCEPTION_ILLEGAL_INSTRUCTION:

       str = "Illegal Instruction";
       break;

  case EXCEPTION_NONCONTINUABLE_EXCEPTION:

       str = "Non-continuable Exception";
       break;

  case EXCEPTION_STACK_OVERFLOW:

       str = "Stack Overflow";
       break;

  case EXCEPTION_INVALID_DISPOSITION:

       str = "Invalid Disposition";
       break;

  case EXCEPTION_GUARD_PAGE:

       str = "Guard Page";
       break;

  case EXCEPTION_INVALID_HANDLE:

       str = "Invalid Handle";
       break;

  default:

       str = "Unknown";
       break;
  }
 
 return str;
}

/*************************************************************************
 This function translates the exception to a displayable string.
*************************************************************************/
std::string Win32Exception::toString() const
{
 char        szAddress[10];
 std::string str;

 str = "An exception of type 0x";

 wsprintf(szAddress, fmt, dwExcCode);
 str.append(szAddress);

 str.append( " - '");
 str.append( getType() );
 str.append( "' - occurred at 0x");

 wsprintf(szAddress, fmt, dwAddress );
 str.append(szAddress);

 str.append( " in module '");
 str.append( getModuleName() );
 str.append( "' at offset 0x");
 
 wsprintf(szAddress, fmt, getModuleOffset() );
 str.append(szAddress);

 return str;
}

/*************************************************************************
 This function returns the name of the file that contains the module 
 where the exception occurred.
*************************************************************************/
std::string Win32Exception::getModuleName() const
{
 int                      n;
 char                     szModule[_MAX_PATH + 1];
 std::string              str;
 MEMORY_BASIC_INFORMATION mbi;

 ZeroMemory(&mbi, sizeof(mbi));

 VirtualQuery(reinterpret_cast<const void *>(dwAddress), &mbi, sizeof(mbi));

 n = GetModuleFileName( reinterpret_cast<HINSTANCE>(mbi.AllocationBase), 
                        szModule, _MAX_PATH);
 if ( n > 0 )
  str = szModule;

 else
  str = "Unknown?";

 return str;
}

/*************************************************************************
 This function returns the offset from the start of the module to 
 the instruction where the exception occurred.
*************************************************************************/
DWORD Win32Exception::getModuleOffset() const
{
 DWORD                    dwOffset;
 MEMORY_BASIC_INFORMATION mbi;

 ZeroMemory(&mbi, sizeof(mbi));

 VirtualQuery( reinterpret_cast<const void *>(dwAddress), &mbi, sizeof(mbi));

 dwOffset = dwAddress - reinterpret_cast<DWORD>(mbi.AllocationBase);

 return dwOffset;
}

